unit Graph;

interface

uses
  Windows, SysUtils, Classes, Controls, Graphics, My_Consts, Data;

type
  TToneGraphicType = (ttNormalUp, ttHighUp, ttNormalDown, ttHighDown, ttBreak);

  TGraph = class(TCustomControl)
  private
    FData: TData;
    FBitmap: TBitmap;
    FKeyBitmap: TBitmap;
    FMeasureBitmaps: array[TMeasure] of TBitmap;
    FTonesBitmaps: array[TToneGraphicType, 1..6] of TBitmap;

    procedure DrawLines;
    procedure DrawKey;
    procedure DrawMeasure;
    procedure DrawTone(Index: integer);
    procedure DrawSeparators;
    procedure DrawTones;

    procedure LoadBitmaps;
    procedure UnloadBitmaps;

  protected
    procedure CreateWnd; override;
    procedure Paint; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property Data: TData read FData write FData;

    procedure Update; override;
  end;

implementation

constructor TGraph.Create(AOwner: TComponent);
begin
  inherited;

  FBitmap := TBitmap.Create;
  with FBitmap do
  begin
    Height := 80;
    Width := 1000;
  end;

  LoadBitmaps;
end;

destructor TGraph.Destroy;
begin
  UnloadBitmaps;
  FBitmap.Free;
  inherited;
end;

procedure TGraph.DrawLines;
var
  i: integer;

begin
  FBitmap.Canvas.Pen.Color := clBlack;

  for i := 0 to 4 do
    with FBitmap.Canvas do
    begin
      MoveTo(0, LinesOffsetY + i * HeightOfRow);
      LineTo(self.Width, LinesOffsetY + i * HeightOfRow);
    end;
end;

procedure TGraph.Paint;
begin
  BitBlt(Canvas.Handle, 0, 0, Width, Height, FBitmap.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TGraph.CreateWnd;
begin
  inherited;
end;

procedure TGraph.DrawKey;
begin
  BitBlt(FBitmap.Canvas.Handle, KeyOffsetX, KeyOffsetY, FKeyBitmap.Width, FKeyBitmap.Height,
    FKeyBitmap.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TGraph.DrawMeasure;
var
  Src: HDC;
  w, h: integer;

begin
  if FData = nil then
    exit;

  with FMeasureBitmaps[FData.Measure] do
  begin
    Src := Canvas.Handle;
    w := Width;
    h := Height;
  end;

  BitBlt(FBitmap.Canvas.Handle, MeasureOffsetX, MeasureOffsetY, w, h, Src, 0, 0, SRCCOPY);
end;

const
  TonesCenterPoints: array[TToneGraphicType, 1..6] of TPoint = (
    ((X: 4; Y: 21), (X: 4; Y: 21), (X: 4; Y: 21), (X: 4; Y: 21), (X: 4; Y: 21), (X: 6; Y: 3)),
    ((X: 12; Y: 21), (X: 12; Y: 21), (X: 12; Y: 21), (X: 12; Y: 21), (X: 12; Y: 21), (X: 15; Y: 8)),
    ((X: 3; Y: 3), (X: 3; Y: 3), (X: 3; Y: 3), (X: 3; Y: 3), (X: 3; Y: 3), (X: 6; Y: 3)),
    ((X: 12; Y: 9), (X: 12; Y: 9), (X: 12; Y: 9), (X: 12; Y: 9), (X: 12; Y: 9), (X: 15; Y: 8)),
    ((X: 0; Y: -10), (X: 0; Y: -5), (X: 0; Y: -5), (X: 10; Y: -13), (X: 10; Y: -13), (X: 8; Y: -8))
  );

procedure TGraph.DrawTone(Index: integer);
var
  ax, ay, cx, cy: integer;
  Entry: PTone;
  b: TBitmap;
  t: TToneGraphicType;

begin
  if Data = nil then
    exit;

  Entry := Data.GetTone(Index);
  if Entry = nil then
    exit;

  if Entry^.Freq = 0 then
    t := ttBreak
  else
    if Entry^.Freq in NormalUpTones then
      t := ttNormalUp
    else
      if Entry^.Freq in HighUpTones then
        t := ttHighUp
      else
        if Entry^.Freq in NormalDownTones then
          t := ttNormalDown
        else
          if Entry^.Freq in HighDownTones then
            t := ttHighDown
          else
            exit;

  b := FTonesBitmaps[t, Entry^.Length];

  cx := TonesOffsetX + Index * TonesSpaceX;
  cy := TonesOffsetsY[Entry^.Freq] + LinesOffsetY;

  ax := cx - TonesCenterPoints[t, Entry^.Length].X;
  ay := cy - TonesCenterPoints[t, Entry^.Length].Y;

  BitBlt(FBitmap.Canvas.Handle, ax, ay, b.Width, b.Height, b.Canvas.Handle, 0, 0, SRCCOPY);

  if Entry^.Freq in LowLineTones then
  begin
    FBitmap.Canvas.MoveTo(cx-7, LowLineOffsetY);
    FBitmap.Canvas.LineTo(cx+8, LowLineOffsetY);
  end;

  if Entry^.Freq in HighLineTones then
  begin
    FBitmap.Canvas.MoveTo(cx-7, HighLineOffsetY);
    FBitmap.Canvas.LineTo(cx+8, HighLineOffsetY);
  end;
end;

procedure TGraph.DrawSeparators;
const
  LengthToReal: array[1..6] of integer = (5, 10, 15, 20, 30, 40);
  MeasureToCount: array[TMeasure] of integer = (30, 40, 50);

var
  i, x: integer;
  sum: integer;
  Entry: PTone;

begin
  if Data = nil then exit;

  sum := 0;

  for i := 0 to Data.Count - 1 do
  begin
    Entry := Data.GetTone(i);
    sum := sum + LengthToReal[Entry^.Length];
    if (sum mod MeasureToCount[Data.Measure]) = 0 then
    begin
      x := Round((i + 0.5) * TonesSpaceX) + TonesOffsetX;

      FBitmap.Canvas.MoveTo(x, LinesOffsetY);
      FBitmap.Canvas.LineTo(x, LinesOffsetY + 4 * HeightOfRow);
    end;
  end;

end;


procedure TGraph.DrawTones;
const
  LengthToReal: array[1..6] of integer = (5, 10, 15, 20, 30, 40);
  MeasureToCount: array[TMeasure] of integer = (30, 40, 50);

var
  i, x, sum: integer;
  ax, ay, cx, cy: integer;
  Entry: PTone;
  b: TBitmap;
  t: TToneGraphicType;

begin
  if Data = nil then
    exit;

  x := TonesOffsetX;
  sum := 0;

  FBitmap.Width := x + TonesOffsetX;

  for i := 0 to Data.Count - 1 do
  begin
    Entry := Data.GetTone(i);
    if Entry = nil then
      exit;

    if Entry^.Freq = 0 then
      t := ttBreak
    else
      if Entry^.Freq in NormalUpTones then
        t := ttNormalUp
      else
        if Entry^.Freq in HighUpTones then
          t := ttHighUp
        else
          if Entry^.Freq in NormalDownTones then
            t := ttNormalDown
          else
            if Entry^.Freq in HighDownTones then
              t := ttHighDown
            else
              exit;

    b := FTonesBitmaps[t, Entry^.Length];

    cx := x;
    cy := TonesOffsetsY[Entry^.Freq] + LinesOffsetY;

    ax := cx - TonesCenterPoints[t, Entry^.Length].X;
    ay := cy - TonesCenterPoints[t, Entry^.Length].Y;

    BitBlt(FBitmap.Canvas.Handle, ax, ay, b.Width, b.Height, b.Canvas.Handle, 0, 0, SRCCOPY);

    if Entry^.Freq in LowLineTones then
    begin
      FBitmap.Canvas.MoveTo(cx-7, LowLineOffsetY);
      FBitmap.Canvas.LineTo(cx+8, LowLineOffsetY);
    end;

    if Entry^.Freq in HighLineTones then
    begin
      FBitmap.Canvas.MoveTo(cx-7, HighLineOffsetY);
      FBitmap.Canvas.LineTo(cx+8, HighLineOffsetY);
    end;

    x := x + TonesSpaceX;

    sum := sum + LengthToReal[Entry^.Length];
    if (sum mod MeasureToCount[Data.Measure]) = 0 then
    begin
      FBitmap.Canvas.MoveTo(x, LinesOffsetY);
      FBitmap.Canvas.LineTo(x, LinesOffsetY + 4 * HeightOfRow);
      x := x + TonesSpaceX;
    end;

    FBitmap.Width := x + TonesOffsetX;
  end;

  Width := FBitmap.Width;
end;


procedure TGraph.LoadBitmaps;

  function LoadBitmap(AName: string): TBitmap;
  begin
    result := TBitmap.Create;
    result.LoadFromResourceName(HInstance, AName);
  end;

const
  ResNames: array[TToneGraphicType] of string = ('UP_NORMAL', 'UP_HIGH', 'DOWN_NORMAL', 'DOWN_HIGH', 'BREAK');

var
  t: TToneGraphicType;
  i: integer;
  s: string;

begin
  FKeyBitmap := LoadBitmap('KEY');
  FMeasureBitmaps[mt34] := LoadBitmap('MEASURE_34');
  FMeasureBitmaps[mt44] := LoadBitmap('MEASURE_44');
  FMeasureBitmaps[mt54] := LoadBitmap('MEASURE_54');

  for t := Low(TToneGraphicType) to High(TToneGraphicType) do
  begin
    for i := 1 to 6 do
    begin
      s := 'TONE_' + ResNames[t] + '_' + IntToStr(i);
      FTonesBitmaps[t, i] := LoadBitmap(s);
    end;
  end;
end;

procedure TGraph.UnloadBitmaps;
var
  m: TMeasure;

begin
  FKeyBitmap.Free;
  for m := Low(m) to High(m) do FMeasureBitmaps[m].Free;
end;

procedure TGraph.Update;
var
  i: integer;

begin
  FBitmap.Canvas.Brush.Color := clWhite;
  FBitmap.Canvas.FillRect(Bounds(0, 0, FBitmap.Width, FBitmap.Height));

  DrawKey;
  DrawMeasure;
  DrawTones;
  DrawLines;

  Invalidate;
end;

end.
